__all__ = (  # noqa: F405
    "build_python_env",
    "clean_pycache_dirs",
    "replace_absolute_paths",
    "search_replace_in_files",
    "search_replace_in_file",
)

from .build import (
    build_python_env,
    clean_pycache_dirs,
    replace_absolute_paths,
)
from .search_replace_files import search_replace_in_file, search_replace_in_files



"""
py-app-standalone uses uv to build a standalone, relocatable Python installation with
the given packages installed.

More info: https://github.com/jlevy/py-app-standalone
"""

import argparse
import shutil
import subprocess
from importlib.metadata import version
from pathlib import Path
from typing import Any

from rich import get_console
from rich_argparse.contrib import ParagraphRichHelpFormatter

from py_app_standalone.build import build_python_env
from py_app_standalone.cli_utils import fail

APP_NAME = "py-app-standalone"

DEFAULT_PYTHON_VERSION = "3.13"

DEFAULT_TARGET_DIR = "py-standalone"


def get_app_version() -> str:
    try:
        return "v" + version(APP_NAME)
    except Exception:
        return "unknown"


def main() -> int:
    """
    Main entry point for the CLI.
    """
    parser = build_parser()
    args = parser.parse_args()

    if not shutil.which("uv"):
        fail("Could not find uv. Follow instructions to install at: https://docs.astral.sh/uv/")

    try:
        build_python_env(
            args.packages,
            args.target,
            args.python_version,
            source_only=args.source_only,
            force=args.force,
        )
    except (FileNotFoundError, FileExistsError, subprocess.CalledProcessError) as e:
        fail(f"{e}")

    return 0


def build_parser() -> argparse.ArgumentParser:
    class CustomFormatter(ParagraphRichHelpFormatter):
        def __init__(self, *args: Any, **kwargs: Any) -> None:
            width = max(40, min(88, get_console().width))
            super().__init__(*args, width=width, **kwargs)

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=CustomFormatter,
    )

    parser.add_argument("--version", action="version", version=f"{APP_NAME} {get_app_version()}")

    parser.add_argument(
        "packages",
        type=str,
        nargs="+",
        help="Packages to install in the virtual environment, in the same format as to pip.",
    )

    parser.add_argument(
        "--source-only",
        action="store_true",
        help="Only install source .py files, ensure no .pyc files are included. Without this flag, "
        "all sources are also compiled.",
    )

    parser.add_argument(
        "--target",
        type=Path,
        default=DEFAULT_TARGET_DIR,
        help="Target directory for the virtual environment, in the format used "
        f"by uv. Default: {DEFAULT_TARGET_DIR}",
    )

    parser.add_argument(
        "--python-version",
        type=str,
        default=DEFAULT_PYTHON_VERSION,
        help=f"Python version to use for the virtual environment. Default: {DEFAULT_PYTHON_VERSION}",
    )

    parser.add_argument(
        "--force",
        action="store_true",
        help="Force overwrite of existing target directory.",
    )

    return parser


if __name__ == "__main__":
    main()



import compileall
import glob
import os
import shutil
import sys
from pathlib import Path

from prettyfmt import fmt_path

from py_app_standalone.cli_utils import fail, info, run, success
from py_app_standalone.search_replace_files import search_replace_in_files
from py_app_standalone.shebangs import RELOCATABLE_PYTHON3_SHEBANG, replace_shebangs


def build_python_env(
    package_list: list[str],
    target_dir: Path,
    python_version: str,
    source_only: bool = False,
    force: bool = False,
):
    """
    Use uv to create a standalone Python environment with the given module installed.
    Packages can be listed as they would be to pip.
    """

    # This is the absolute path to the venv, which will be used by pip.
    target_absolute = target_dir.resolve()

    if target_dir.exists() and not force:
        raise FileExistsError(
            f"target directory already exists (run with --force to run anyway): {target_dir}"
        )

    info(f"Creating a new Python installation at: {fmt_path(target_absolute)}")
    run(
        [
            "uv",
            "python",
            "install",
            # These are the uv-managed standalone Python binaries.
            "--managed-python",
            "--install-dir",
            str(target_absolute),
            python_version,
        ]
    )

    # Find the root directory of the environment.
    install_root_pat = os.path.join(target_dir, f"cpython-{python_version}.*")
    install_root_paths = glob.glob(install_root_pat)
    if not install_root_paths:
        raise FileNotFoundError(f"Failed to find venv root at: {install_root_pat}")

    install_root = Path(install_root_paths[0])  # Use the first match

    # XXX There is currently no option to run `uv pip install --relocatable` but
    # there is an option to do it with `uv venv --relocatable`.
    # So let's get a config file that marks a venv as relocatable.
    bare_venv_dir = target_dir / "bare-venv"
    temp_pyvenv_cfg = install_root / "pyvenv.cfg"
    info(
        "Creating temporary venv (you can ignore this step; it is just a trick "
        "to get a uv pyvenv.cfg file)..."
    )
    run(
        [
            "uv",
            "venv",
            "--relocatable",
            "--python",
            str(install_root),
            str(bare_venv_dir),
        ]
    )
    os.rename(bare_venv_dir / "pyvenv.cfg", temp_pyvenv_cfg)
    shutil.rmtree(bare_venv_dir)
    info(f"Temporary venv config: {fmt_path(temp_pyvenv_cfg)}")

    info(
        f"Now installing packages directly into the original installation: {fmt_path(install_root)}"
    )
    run(
        [
            "uv",
            "pip",
            "install",
            *package_list,
            "--python",
            str(install_root),
            "--break-system-packages",
        ]
    )

    # Now we don't need the temp pyvenv.cfg file anymore.
    # It has absolute paths so better to just remove it.
    os.remove(temp_pyvenv_cfg)

    # Note we could compile everything like this, but it would add absolute paths within the pyc files:
    # compileall.compile_dir(target_absolute, quiet=1)

    # First handle binaries with possible absolute paths.
    if sys.platform == "darwin":
        info("macOS: Updating dylib ids to be relocatable...")
        update_macos_dylib_ids(install_root)

    # Make all the scripts relocatable.
    # This may not actually be necessary anymore since we've done the pyvenv.cfg hack
    # above, but it's no harm to have it here too.
    info("Making sure all the scripts are relocatable...")
    replace_shebangs([f"{install_root}/bin/*"], RELOCATABLE_PYTHON3_SHEBANG)

    # Then handle text files with absolute paths.
    info("Replacing remaining absolute paths...")
    replace_absolute_paths(install_root, str(target_absolute), str(target_dir))

    if source_only:
        info("Source only build: cleaning .pyc files...")
        clean_pycache_dirs(target_absolute)
    else:
        info(f"Compiling all python files in: {fmt_path(target_absolute)}...")
        # Important to include stripdir to avoid absolute paths.
        compileall.compile_dir(target_absolute, quiet=1, stripdir=target_absolute)

    # Double check there are no absolute paths left.
    sanity_check_absolute_paths(install_root, str(target_absolute))

    success(
        f"Created standalone Python environment for packages {package_list} "
        f"at: {fmt_path(target_dir)}"
    )


def update_macos_dylib_ids(python_root: Path):
    """
    macOS: Update the dylib ids of all the dylibs in the given root directory.
    It seems uv patches them with absolute paths using `install_name_tool`, which means
    they are tied to the user's install location.

    So we reset this but use `@executable_path` to ensure the dylibs are relocatable.
    See:
    https://github.com/astral-sh/uv/blob/main/crates/uv-python/src/managed.rs#L545-L567
    https://github.com/astral-sh/uv/blob/main/crates/uv-python/src/macos_dylib.rs
    """
    # Should be something like: py-standalone/cpython-3.13.2-macos-aarch64-none/lib/libpython3.13.dylib
    glob_pattern = f"{python_root}/lib/**/*.dylib"
    for dylib_path in glob.glob(glob_pattern, recursive=True):
        info(f"Found macos dylib, will update its id to remove any absolute paths: {dylib_path}")
        rel_path = Path(dylib_path).relative_to(python_root)
        run(["install_name_tool", "-id", f"@executable_path/../{rel_path}", dylib_path])


def replace_absolute_paths(python_root: Path, old_path_str: str, new_path_str: str):
    """
    Replace all old (absolute) paths with the new (relative) path.
    This works fine on all text files. We skip binary libs but most do not have absolute paths.
    """
    text_glob = [f"{python_root}/bin/*", f"{python_root}/lib/**/*.py"]

    info()
    info(
        "Replacing all absolute paths in:\n"
        f"    {' '.join(text_glob)}:\n"
        f"    `{old_path_str}` -> `{new_path_str}`"
    )
    matches, _files_changed = search_replace_in_files(
        text_glob, old_path_str.encode(), new_path_str.encode()
    )
    info(f"Replaced {matches} total occurrences in {len(_files_changed)} files total")


def sanity_check_absolute_paths(python_root: Path, old_path_str: str):
    info()
    info("Sanity checking if any absolute paths remain (including in binary files)...")
    all_files_glob = [f"{python_root}/**/*"]
    matches, _files_changed = search_replace_in_files(all_files_glob, old_path_str.encode(), None)
    if matches:
        fail(f"Found {matches} matches of `{old_path_str}` in binary files (see above)")
    else:
        info("Great! No absolute paths found in the installed files.")


def clean_pycache_dirs(dir_path: Path):
    """
    Remove all __pycache__ directories within the given root directory.
    """
    for dirpath, dirnames, _filenames in os.walk(dir_path):
        if "__pycache__" in dirnames:
            pycache_path = os.path.join(dirpath, "__pycache__")
            shutil.rmtree(pycache_path)
            info(f"Removed: {fmt_path(pycache_path)}")



import glob
import os
from pathlib import Path

from prettyfmt import fmt_path
from strif import atomic_output_file

from py_app_standalone.cli_utils import info


def search_replace_in_files(
    glob_patterns: list[str],
    search_bytes: bytes,
    replace_bytes: bytes | None,
    backup_suffix: str | None = None,
) -> tuple[int, list[Path]]:
    """
    Search or replace occurrences of `search_bytes` with `replace_bytes` in all files
    matching the glob pattern.
    """
    files_with_replacements: list[Path] = []
    total_replacements = 0

    for glob_pattern in glob_patterns:
        for file_path in glob.glob(glob_pattern, recursive=True):
            file_path = Path(file_path)
            if file_path.is_file():
                replacements = search_replace_in_file(
                    file_path, search_bytes, replace_bytes, backup_suffix
                )
                if replacements > 0:
                    op = "Replaced" if replace_bytes else "Found"
                    info(f"{op} {replacements} occurrences in: {fmt_path(file_path)}")
                    total_replacements += replacements
                    files_with_replacements.append(file_path)
    return total_replacements, files_with_replacements


def search_replace_in_file(
    path: Path,
    search_bytes: bytes,
    replace_bytes: bytes | None,
    backup_suffix: str | None = None,
) -> int:
    """
    Search or replace occurrences of `search_bytes` with `replace_bytes` in a file,
    atomically and in place.
    Reads entire file into memory!
    Returns the number of matches found or replacements made.
    """
    with path.open("rb") as file:
        content = file.read()

    replacements = content.count(search_bytes)
    if replacements > 0 and replace_bytes:
        original_permissions = os.stat(path).st_mode

        content = content.replace(search_bytes, replace_bytes)
        with atomic_output_file(path, backup_suffix=backup_suffix) as file:
            with file.open("wb") as f:
                f.write(content)

        # Restore the original file permissions.
        os.chmod(path, original_permissions)

    return replacements



import glob
import os
from pathlib import Path

from prettyfmt import fmt_path
from strif import atomic_output_file

from py_app_standalone.cli_utils import info

# Relocatable shebang for posix.
# This is what uv does, which is based on pip's distlib. See:
# https://github.com/astral-sh/uv/blob/main/crates/uv-install-wheel/src/wheel.rs#L105-L140
# https://github.com/pypa/pip/blob/0ad4c94be74cc24874c6feb5bb3c2152c398a18e/src/pip/_vendor/distlib/scripts.py#L136-L165
RELOCATABLE_PYTHON3_SHEBANG = """
#!/bin/sh
'''exec' "$(dirname -- "$(realpath -- "$0")")"/'python3' "$0" "$@"
' '''
""".lstrip()


def add_shebang(
    path: Path,
    new_shebang: str,
    backup_suffix: str | None = None,
) -> str | None:
    """
    Replace the shebang in a script file and return the old shebang if present.
    Handles both single-line and multi-line shebangs.
    """
    if not path.is_file():
        return None

    with path.open("rb") as file:
        content = file.read()

    # Split into lines with line endings preserved.
    lines = content.splitlines(True)

    # Check if there's a shebang.
    if not lines or not lines[0].startswith(b"#!"):
        return None

    # Determine if it's a multi-line shebang.
    is_multiline = len(lines) >= 3 and lines[1].startswith(b"'") and lines[2].startswith(b"'")

    # Determine how many lines to replace.
    lines_to_replace = 3 if is_multiline else 1

    # Capture the old shebang.
    old_shebang = b"".join(lines[:lines_to_replace]).decode()

    # Combine the new shebang with the rest of the file.
    new_content = new_shebang.encode()
    if lines_to_replace < len(lines):
        new_content += b"".join(lines[lines_to_replace:])

    # Store the original permissions.
    original_permissions = os.stat(path).st_mode

    # Write back using atomic file operations.
    with atomic_output_file(path, backup_suffix=backup_suffix) as file:
        with file.open("wb") as f:
            f.write(new_content)

    # Restore the original file permissions.
    os.chmod(path, original_permissions)

    return old_shebang


def replace_shebangs(
    glob_patterns: list[str],
    new_shebang: str,
    backup_suffix: str | None = None,
) -> int:
    """
    Replace shebangs in all files matching the glob patterns with the new shebang.
    Returns the total number of replacements made.
    """
    total_replacements = 0

    info()
    info("Inserting relocatable shebangs on scripts in:\n" + f"    {', '.join(glob_patterns)}")
    for glob_pattern in glob_patterns:
        for file_path in glob.glob(glob_pattern, recursive=True):
            path = Path(file_path)
            if path.is_file():
                old_shebang = add_shebang(path, new_shebang, backup_suffix)
                if old_shebang is not None:
                    info(f"Replaced shebang in: {fmt_path(file_path)}")
                    total_replacements += 1

    return total_replacements



import glob
import os
import shutil
import subprocess
import sys
from os import chdir
from pathlib import Path

from py_app_standalone import build_python_env


def test_simple_env():
    target_dir = Path("./py-standalone")
    if target_dir.exists():
        shutil.rmtree(target_dir)
    new_python_root = Path("./renamed_root")
    if new_python_root.exists():
        shutil.rmtree(new_python_root)

    build_python_env(["cowsay"], target_dir, "3.13")

    assert target_dir.exists()
    assert target_dir.is_dir()

    python_roots = glob.glob(str(target_dir / "cpython-3.13.*"))
    assert len(python_roots) == 1
    python_root = Path(python_roots[0]).resolve()

    # Determine cowsay binary relative path based on platform
    bin_dir = "Scripts" if sys.platform == "win32" else "bin"
    cowsay_extension = ".exe" if sys.platform == "win32" else ""
    cowsay_rel_path = Path(bin_dir) / f"cowsay{cowsay_extension}"

    # Check for cowsay in the correct location
    assert (python_root / cowsay_rel_path).exists()
    assert (python_root / cowsay_rel_path).is_file()

    def run(cmd: list[str]):
        print(f"\nRunning: {cmd}")
        subprocess.run(cmd, check=True)

    run([str(python_root / cowsay_rel_path), "-t", "Hello, world!"])

    os.rename(python_root, new_python_root)

    run([str(new_python_root / cowsay_rel_path), "-t", "Hello, world from a new folder!"])

    # Confirm that the cwd doesn't matter.
    new_python_root = new_python_root.resolve()

    # Doesn't really matter where, just another dir.
    another_dir = new_python_root.parent.parent
    chdir(another_dir)
    run([str(new_python_root / cowsay_rel_path), "-t", "Hello, world from a different directory!"])



import subprocess
import sys

from funlog import log_calls
from rich import get_console, reconfigure
from rich import print as rprint

reconfigure(emoji=not get_console().options.legacy_windows)  # No emojis on legacy windows.


@log_calls(level="warning", show_timing_only=True)
def run(cmd: list[str]):
    rprint()
    rprint(f"[bold green]:arrow_forward: {' '.join(cmd)}[/bold green]")
    subprocess.run(cmd, text=True, check=True)
    rprint()


def info(msg: str = ""):
    rprint(f"[blue]{msg}[/blue]")


def warn(msg: str):
    rprint()
    rprint(f"[bold yellow]Warning: {msg}[/bold yellow]")
    rprint()


def success(msg: str):
    rprint()
    rprint(f"[bold green]:heavy_check_mark: Success: {msg}[/bold green]")
    rprint()


def fail(msg: str):
    rprint()
    rprint(f"[bold red]:x: Error: {msg}[/bold red]")
    rprint()
    sys.exit(1)
